#!/usr/bin/env python

"""
Functions to interact with the REDCap API.
"""

from collections import OrderedDict
from dataclasses import dataclass, field
from datetime import datetime, timezone
from functools import singledispatch
from json import dump, dumps, loads
from os import environ, getenv, walk
from os.path import join
from pathlib import Path
from typing import Sequence
from warnings import warn
from zipfile import ZIP_DEFLATED, ZipFile

import polars as pl
import pytz
import requests
from dateutil.relativedelta import relativedelta as rdelta
from dotenv import find_dotenv, load_dotenv

from babylab.globals import COLNAMES, FIELDS_TO_RENAME, INT_FIELDS, SCHEMA, URI


class MissingEnvFile(Exception):
    """.env file is not found in user folder"""


class MissingEnvToken(Exception):
    """Token is not provided as key in .env"""


class MissingRecord(Exception):
    """Record is not found"""


class BadToken(Exception):
    """Token is ill-formed"""


class BadAgeFormat(Exception):
    """If age does not follow the right format"""


@dataclass
class RecordList:
    """List of REDCap records."""

    records: dict = field(default_factory=dict)
    kind: str | None = None

    def __len__(self) -> int:
        return len(self.records)


@dataclass
class Record:
    ppt_id: str
    data: dict


@dataclass
class Participant(Record):
    appointments: RecordList = field(default_factory=list)
    questionnaires: RecordList = field(default_factory=list)


@dataclass
class Appointment(Record):
    def __post_init__(self):
        apt_id = self.data["redcap_repeat_instance"]
        self.apt_id = make_id(self.ppt_id, apt_id)
        self.status: str = self.data["status"]
        self.date: str = parse_str_date(self.data["date"])


@dataclass
class Questionnaire(Record):
    def __post_init__(self):
        que_id = self.data["redcap_repeat_instance"]
        self.que_id = make_id(self.ppt_id, que_id)
        self.isestimated = self.data["isestimated"]


def get_api_key(path: Path | str | None = None, name: str = "API_KEY") -> str:
    """Retrieve API credentials.

    Args:
        path (Path | str | None, optional): Path to the .env file with global variables. Defaults to ``Path.home()``.
        name (str, optional): Name of the variable to import. Defaults to "API_KEY".

    Returns:
        str: API key token.

    Raises:
        MissingEnvFile: If .env file is not found  in ``path``.
        MissingEnvToken: If requested environmental variable key is not found.
        BadToken: If token contains any non-alphanumeric character.
    """
    if name in environ or getenv("GITHUB_ACTIONS") == "true":
        token = getenv(name)
    else:
        path = Path(find_dotenv()) if not path else Path(path)

        if not path.exists():
            raise MissingEnvFile(f".env file not found in {path}")

        load_dotenv(path, override=True)
        token = getenv(name)

    if token is None:
        raise MissingEnvToken(f"No environment variable named '{name}' found")

    if not isinstance(token, str) or not token.isalnum():
        raise BadToken("Token must be str with no non-alphanumeric characters")

    return token


def post_request(fields: dict, timeout: Sequence[int] = (5, 10)) -> requests.Response:
    """Make a POST request to the REDCap database.

    Args:
        fields (dict): Fields to retrieve.
        timeout (Sequence[int], optional): Timeout of HTTP request in seconds. Defaults to 10.

    Raises:
        requests.exceptions.HTTPError: If HTTP request fails.
        BadToken: If API token contains non-alphanumeric characters.

    Returns:
        requests.Response: HTTP request response in JSON format.
    """
    t = get_api_key()

    if t is None:
        raise MissingEnvToken("No key found in your .env file")

    fields = OrderedDict(fields)
    fields["token"] = t
    fields.move_to_end("token", last=False)

    r = requests.post(URI, data=fields, timeout=timeout)
    r.raise_for_status()

    return r


def get_data_dict() -> dict:
    """Get data dictionaries for categorical variables.

    Returns:
        dict: Data dictionary.
    """
    fields = {"content": "metadata", "format": "json", "returnFormat": "json"}

    for idx, i in enumerate(FIELDS_TO_RENAME):
        fields[f"fields[{idx}]"] = i

    r = loads(post_request(fields=fields).text)
    items_ordered = [i["field_name"] for i in r]
    dicts = {}

    for k, v in zip(items_ordered, r, strict=False):
        options = v["select_choices_or_calculations"].split("|")
        options = [tuple(o.strip().split(", ")) for o in options]

        if k.startswith("language_"):
            options = sorted(options, key=lambda x: x[1])

        dicts[k] = dict(options)

    return dicts


DATA_DICT: dict = get_data_dict()


def get_redcap_version() -> str:
    """Get REDCap version.

    Returns:
        str: REDCAp version number.
    """
    fields = {"content": "version"}
    r = post_request(fields=fields)

    return r.content.decode("utf-8")


def to_df(x: RecordList) -> pl.DataFrame:
    """Transforms a RecordList to a Polars DataFrame.

    Returns:
        pl.DataFrame: Tabular data frame.
    """
    recs = [p.data for p in x.records.values()]
    names = COLNAMES[x.kind]

    if not recs:
        return pl.DataFrame(schema=names)

    id_lookup = {
        "participants": "none",
        "appointments": "apt_id",
        "questionnaires": "que_id",
    }

    df = (
        pl.DataFrame(recs, schema=SCHEMA[x.kind])
        .rename({"redcap_repeat_instance": id_lookup[x.kind]}, strict=False)
        .with_columns(
            pl.when(pl.col(pl.String).str.len_chars() == 0)
            .then(None)
            .otherwise(pl.col(pl.String))
            .name.keep()
        )
        .with_columns(
            pl.col(
                [f for f in INT_FIELDS if f in names],
            ).cast(pl.Int128)
        )
    )

    return df


def filter_fields(data: dict, prefix: str, fields: list[str]) -> dict:
    """Filter a data dictionary based on a prefix and field names.

    Args:
        records (dict): Record data dictionary.
        prefix (str): Prefix to look for.
        fields (list[str]): Field names to look for.

    Returns:
        dict: Filtered records.
    """
    return {
        k.replace(prefix, ""): v
        for k, v in data.items()
        if k.startswith(prefix) or k in fields
    }


@singledispatch
def fmt_labels(x: dict | pl.DataFrame):
    """Reformat dataframe.

    Args:
        x (dict | DataFrame): Dataframe to reformat.
        prefixes (list[str]): List of prefixes to look for in variable names.

    Returns:
        DataFrame: A reformated Dataframe.
    """
    raise TypeError("`x` must be a dict or a DataFrame")


@fmt_labels.register(dict)
def _(x: dict) -> dict:
    """Reformat dictionary.

    Args:
        x (dict): dictionary to reformat.

    Returns:
        dict: A reformatted dictionary.
    """
    fields = ["participant_", "appointment_", "language_"]
    y = dict(x)

    for k, v in y.items():
        for f in fields:
            if f + k in DATA_DICT and v:
                y[k] = DATA_DICT[f + k][v]

        if "exp" in k:
            y[k] = round(float(v), None) if v else None

        for c in ["taxi_isbooked", "isdropout", "isestimated"]:
            if c in k:
                y[k] = y[c] == "1"

        y[k] = y[k] if y[k] != "" else None

    y = {k: (int(v) if v and k in INT_FIELDS else v) for k, v in y.items()}

    return y


@fmt_labels.register(pl.DataFrame)
def _(x: pl.DataFrame) -> pl.DataFrame:
    """Reformat DataFrame.

    Args:
        x (dict): dictionary to reformat.

    Returns:
        DataFrame: A reformatted DataFrame.
    """
    cols = {k.rsplit("_", 1)[1]: v for k, v in DATA_DICT.items()}

    for k, v in {ck: cv for ck, cv in cols.items() if ck in x.columns}.items():
        x = x.with_columns(pl.col(k).replace_strict(v, default=None))

    for c in ["isestimated", "isdropout"]:
        if c in x.columns:
            x = x.with_columns(pl.col(c).eq("1"))

    x = x.with_columns(
        pl.when(pl.col(pl.String).str.len_chars() == 0)
        .then(None)
        .otherwise(pl.col(pl.String))
        .name.keep()
    ).cast({c: pl.Int64 for c in [f for f in INT_FIELDS if f in x.columns]})

    return x


def str_to_dt(data: dict) -> dict:
    """Parse strings in a dictionary as formatted datetimes.

    It first tries to format the date as "Y-m-d H:M:S". If error, it assumes the "Y-m-d H:M" is due and tries to format it accordingly.

    Args:
        data (dict): Dictionary that may contain string formatted datetimes.

    Returns:
        dict: Dictionary with strings parsed as datetimes.
    """
    for k, v in data.items():
        if v and "date" in k:
            try:
                data[k] = datetime.strptime(data[k], "%Y-%m-%d %H:%M:%S")
            except ValueError:
                data[k] = datetime.strptime(data[k], "%Y-%m-%d %H:%M")

    return data


def dt_to_str(data: dict) -> dict:
    """Format datatimes in a dictionary as strings following the ISO 8061 date format.

    Args:
        data (dict): Dictionary that may contain datetimes.

    Returns:
        dict: Dictionary with datetimes formatted as strings.
    """
    for k, v in data.items():
        if isinstance(v, datetime):
            data[k] = data[k].isoformat()

    return data


def get_next_id() -> str:
    """Get next record_id in REDCap database.

    Returns:
        str: record_id of next record.
    """
    fields = {"content": "generateNextRecordName"}

    return str(post_request(fields=fields).json())


def prepare_data(x: dict, kind: str = "ppt") -> dict:
    """Prepare data for class.

    Args:
        x (dict): Participant data retrieved from REDCap.
        kind (str): Type of data to process. Takes "ppt" (default), "apt" or "que".

    Returns:
        dict: Formatted data.

    Raises:
        ValueError: If ``kind`` is not one of 'ppt', 'apt', or 'que'.
    """
    if kind not in ["ppt", "apt", "que"]:
        raise ValueError("`kind` must be one of 'ppt', 'apt', or 'que'")

    if kind == "apt":
        names = ["record_id", "redcap_repeat_instance"]
        x = filter_fields(x, "appointment_", names)

    if kind == "que":
        names = ["record_id", "redcap_repeat_instance"]
        x = filter_fields(x, "language_", names)
        for i in range(1, 5):
            lang = f"lang{i}_exp"
            x[lang] = int(x[lang]) if x[lang] else 0

    if kind == "ppt":
        names = ["record_id"]
        x = filter_fields(x, "participant_", names)
        age_created = (x["age_created_months"], x["age_created_days"])
        months, days = get_age(age_created, x["date_created"])
        x["age_now_months"], x["age_now_days"] = months, days

    return fmt_labels(x)


def make_id(ppt_id: str | int, repeat_id: str | int | None = None) -> str:
    """Make a record ID.

    Args:
        ppt_id (str | int): Participant ID.
        repeat_id (str | int | None, optional): Appointment or Questionnaire ID, or ``redcap_repeated_id``. Defaults to None.

    Returns:
        str: Record ID.
    """
    ppt_id = str(ppt_id)

    if not ppt_id.isdigit():
        raise ValueError(f"`ppt_id`` must be a digit, but '{ppt_id}' was provided")

    if not repeat_id:
        return ppt_id

    repeat_id = str(repeat_id)

    if not repeat_id.isdigit():
        raise ValueError(
            f"`repeat_id`` must be a digit, but '{repeat_id}' was provided"
        )

    return ppt_id + ":" + repeat_id


def get_records(record_id: str | list | None = None) -> dict:
    """Return records as JSON.

    Args:
        record_id  (str): ID of record to retrieve. Defaults to None.

    Returns:
        list[dict[str, str]]: REDCap records in JSON format.
    """
    fields = {"content": "record", "format": "json", "type": "flat"}

    if record_id and isinstance(record_id, list):
        fields["records[0]"] = record_id

        for r in record_id:
            fields[f"records[{record_id}]"] = r

    return post_request(fields=fields).json()


def get_participant(ppt_id: str) -> Participant:
    """Get participant record.

    Args:
        ppt_id: ID of participant (record_id).

    Returns:
        Participant: Participant object.

    Raises:
        MissingRecord: If requested recording is missing in the database.
    """
    fields = {
        "content": "record",
        "action": "export",
        "format": "json",
        "type": "flat",
        "csvDelimiter": "",
        "records[0]": ppt_id,
        "rawOrLabel": "raw",
        "rawOrLabelHeaders": "raw",
        "exportCheckboxLabel": "false",
        "exportSurveyFields": "false",
        "exportDataAccessGroups": "false",
        "returnFormat": "json",
    }

    for i, f in enumerate(["participants", "appointments", "language"]):
        fields[f"forms[{i}]"] = f

    recs = [str_to_dt(r) for r in post_request(fields).json()]
    apt, que = {}, {}

    for r in recs:
        ppt_id = r["record_id"]
        repeat_id = make_id(ppt_id, r["redcap_repeat_instance"])

        if r["redcap_repeat_instrument"] == "appointments":
            data = prepare_data(r, "apt")
            apt[repeat_id] = Appointment(ppt_id=ppt_id, data=data)

        if r["redcap_repeat_instrument"] == "language":
            data = prepare_data(r, "que")
            que[repeat_id] = Questionnaire(ppt_id=ppt_id, data=data)

    try:
        data = prepare_data(recs[0])

        return Participant(
            ppt_id=data["record_id"],
            data=data,
            appointments=RecordList(apt, kind="appointments"),
            questionnaires=RecordList(que, kind="questionnaires"),
        )
    except IndexError as e:
        raise MissingRecord(f"Record {ppt_id} not found") from e


def get_appointment(apt_id: str) -> Appointment:
    """Get appointment record.

    Args:
        apt_id (str): ID of appointment (``redcap_repeated_id``).

    Returns:
        Appointment: Appointment object.

    Raises:
        MissingRecord: If requested record is missing from database.
    """
    ppt_id, _ = apt_id.split(":")
    ppt = get_participant(ppt_id)

    try:
        return ppt.appointments.records[apt_id]
    except KeyError as e:
        raise MissingRecord(f"Record {apt_id} not found") from e


def get_questionnaire(que_id: str) -> Questionnaire:
    """Get questionnaire record.

    Args:
        que_id (str): ID of appointment (``redcap_repeated_id``).

    Returns:
        Questionnaire: Appointment object.
    """
    ppt_id, _ = que_id.split(":")
    ppt = get_participant(ppt_id)

    try:
        return ppt.questionnaires.records[que_id]
    except KeyError as e:
        raise MissingRecord(f"Record {que_id} not found") from e


def add_participant(data: dict, modifying: bool = False):
    """Add new participant to REDCap database.

    Args:
        data (dict): Participant data.
        modifying (bool, optional): Modifying existent participant?
    """
    fields = {
        "content": "record",
        "action": "import",
        "format": "json",
        "type": "flat",
        "overwriteBehavior": "normal" if modifying else "overwrite",
        "forceAutoNumber": "false" if modifying else "true",
        "data": f"[{dumps(dt_to_str(data))}]",
    }

    return post_request(fields=fields)


def delete_participant(data: dict):
    """Delete participant from REDCap database.

    Args:
        data (dict): Participant data.
        modifying (bool, optional): Modifying existent participant?
    """
    fields = {
        "content": "record",
        "action": "delete",
        "returnFormat": "json",
        "instrument": "",
        "records[0]": f"{data['record_id']}",
    }

    r = post_request(fields=fields)
    try:
        r.raise_for_status()
        return r
    except requests.exceptions.HTTPError as e:
        rid = make_id(data["record_id"])
        raise MissingRecord(f"Record {rid} not found") from e


def add_appointment(data: dict):
    """Add new appointment to REDCap database.

    Args:
        record_id (dict): ID of participant.
        data (dict): Appointment data.
    """
    fields = {
        "content": "record",
        "action": "import",
        "format": "json",
        "type": "flat",
        "overwriteBehavior": "overwrite",
        "forceAutoNumber": "false",
        "data": f"[{dumps(dt_to_str(data))}]",
    }

    return post_request(fields=fields)


def delete_appointment(data: dict):
    """Delete appointment from REDCap database.

    Args:
        data (dict): Participant data.
        modifying (bool, optional): Modifying existent participant?
    """
    fields = {
        "content": "record",
        "action": "delete",
        "returnFormat": "json",
        "instrument": "appointments",
        "repeat_instance": int(data["redcap_repeat_instance"]),
        f"records[{data['record_id']}]": f"{data['record_id']}",
    }

    r = post_request(fields=fields)
    warn_missing_record(r)

    return r


def add_questionnaire(data: dict):
    """Add new questionnaire to REDCap database.

    Args:
        data (dict): Questionnaire data.
    """
    fields = {
        "content": "record",
        "action": "import",
        "format": "json",
        "type": "flat",
        "overwriteBehavior": "overwrite",
        "forceAutoNumber": "false",
        "data": f"[{dumps(dt_to_str(data))}]",
    }

    return post_request(fields=fields)


def delete_questionnaire(data: dict):
    """Delete questionnaire from REDCap database.

    Args:
        data (dict): Participant data.
        modifying (bool, optional): Modifying existent participant?
    """
    fields = {
        "content": "record",
        "action": "delete",
        "returnFormat": "json",
        "instrument": "language",
        "repeat_instance": int(data["redcap_repeat_instance"]),
        f"records[{data['record_id']}]": f"{data['record_id']}",
    }

    r = post_request(fields=fields)
    warn_missing_record(r)
    return r


def warn_missing_record(r: requests.models.Response):
    """Warn user about absent record.

    Args:
        r (requests.models.Response): HTTPS response.
    """
    if "registros proporcionados no existen" in r.content.decode():
        warn("Record does not exist!", stacklevel=2)


def redcap_backup(path: Path | str = Path("tmp")) -> Path:
    """Download a backup of the REDCap database

    Args:
        path (Path | str, optional): Output directory. Defaults to ``Path("tmp")``.

    Returns:
        Path: Path to the generated file with data and metadata of the project.
    """
    path = Path(path)
    path.mkdir(exist_ok=True)

    p = {}
    for k in ["project", "metadata", "instrument"]:
        p[k] = {"format": "json", "returnFormat": "json", "content": k}

    d = {k: loads(post_request(v).text) for k, v in p.items()}

    with open(path / "records.csv", "w+", encoding="utf-8") as f:
        fields = {
            "content": "record",
            "action": "export",
            "format": "csv",
            "csvDelimiter": ",",
            "returnFormat": "json",
        }
        records = post_request(fields).content.decode().split("\n")
        records = [r + "\n" for r in records]
        f.writelines(records)

    b = {
        "project": d["project"],
        "instruments": d["instrument"],
        "fields": d["metadata"],
    }

    for k, v in b.items():
        with open(path / (k + ".json"), "w", encoding="utf-8") as f:
            dump(v, f)

    timestamp = datetime.strftime(datetime.now(), "%Y-%m-%d-%H-%M")
    file = path / ("backup_" + timestamp + ".zip")

    for root, _, files in walk(str(path), topdown=False):
        with ZipFile(file, "w", ZIP_DEFLATED) as z:
            for f in files:
                z.write(join(root, f))

    return file


class Records:
    """REDCap records"""

    def __init__(self, record_id: str | list | None = None):
        records = get_records(record_id)
        records = [str_to_dt(r) for r in records]
        ppt, apt, que = {}, {}, {}

        for r in records:
            ppt_id = r["record_id"]
            repeat_id = r["redcap_repeat_instance"]

            if repeat_id and r["appointment_status"]:
                apt_id = make_id(ppt_id, repeat_id)
                r["appointment_id"] = apt_id
                data = prepare_data(r, "apt")

                apt[apt_id] = Appointment(ppt_id=data["record_id"], data=data)

            if repeat_id and r["language_lang1"]:
                que_id = make_id(ppt_id, repeat_id)
                r["questionnaire_id"] = que_id
                data = prepare_data(r, "que")

                que[que_id] = Questionnaire(ppt_id=ppt_id, data=data)

            if not r["redcap_repeat_instrument"]:
                data = prepare_data(r)
                ppt[ppt_id] = Participant(r["record_id"], data)

        # add appointments and questionnaires to each participant
        for p, v in ppt.items():
            apts = {k: v for k, v in apt.items() if v.ppt_id == p}
            v.appointments = RecordList(apts, kind="appointments")
            ques = {k: v for k, v in que.items() if v.ppt_id == p}
            v.questionnaires = RecordList(ques, kind="questionnaires")

        self.participants = RecordList(ppt, kind="participants")
        self.appointments = RecordList(apt, kind="appointments")
        self.questionnaires = RecordList(que, kind="questionnaires")

    def __repr__(self) -> str:
        return (
            "REDCap database:"
            + f"\n- {len(self.participants.records)} participants"
            + f"\n- {len(self.appointments.records)} appointments"
            + f"\n- {len(self.questionnaires.records)} questionnaires"
        )

    def __str__(self) -> str:
        return (
            "REDCap database:"
            + f"\n- {len(self.participants.records)} participants"
            + f"\n- {len(self.appointments.records)} appointments"
            + f"\n- {len(self.questionnaires.records)} questionnaires"
        )


def parse_age(age: tuple) -> tuple[int, int]:
    """Validate age string or tuple.

    Args:
        age (tuple): Age of the participant as a tuple in the ``(months, days)`` format.

    Raises:
        ValueError: If age is not str or tuple.
        BadAgeFormat: If age is ill-formatted.

    Returns:
        tuple[int, int]: Age of the participant in the ``(months, days)`` format.
    """
    try:
        assert isinstance(age, tuple)
        assert len(age) == 2

        return int(age[0]), int(age[1])
    except AssertionError as e:
        raise BadAgeFormat("age must be in (months, age) format") from e


def parse_str_date(x: str | datetime) -> datetime:
    """Parse string data to datetime.

    Args:
        x (str | datetime): String date to parse.

    Returns:
        datetime: Parsed datetime.
    """
    if isinstance(x, datetime):
        return x

    try:
        return datetime.strptime(x, "%Y-%m-%dT%H:%M:%S")
    except ValueError:
        try:
            return datetime.strptime(x, "%Y-%m-%d %H:%M:%S")
        except ValueError:
            return datetime.strptime(x, "%Y-%m-%d %H:%M")


def get_age(
    age: tuple, ts: datetime | str, ts_new: datetime | None = None, tz: str = "UTC"
):
    """Calculate the age of a person in months and days at a new timestamp.

    Args:
        age (tuple): Age in months and days as a tuple of type (months, days).
        ts (datetime | str): Birth date as ``datetime.datetime`` type.
        ts_new (datetime.datetime | None, optional): Time for which the age is calculated. Defaults to current date (``datetime.datetime.now()``).

    Returns:
        tuple: Age in at ``new_timestamp``.
    """
    tz = pytz.timezone(tz)
    ts = tz.localize(parse_str_date(ts))
    ts_new = datetime.now() if ts_new is None else ts_new
    ts_new = tz.localize(ts_new)

    tdiff = rdelta(ts_new, ts)
    months, days = parse_age(age)
    new_age_months = months + tdiff.years * 12 + tdiff.months
    new_age_days = days + tdiff.days

    if new_age_days >= 30:
        additional_months = new_age_days // 30
        new_age_months += additional_months
        new_age_days %= 30

    return new_age_months, new_age_days


if __name__ == "__main__":
    r = Records()



"""
Fixtures for testing
"""

import os
from datetime import datetime
from random import choice, choices
from string import ascii_lowercase, digits

from babylab import api

IS_GIHTUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"
RECORDS: api.Records = api.Records()
DATA_DICT: dict = api.get_data_dict()


def generate_str(n: int = 7) -> str:
    """Generate random string of ASCII characters.

    Args:
        nchar (int, optional): Number of characters in the string. Defaults to 7.

    Returns:
        str: Random string of characters of length ``n``.
    """
    return "".join(choices(ascii_lowercase + digits, k=n))


def generate_phone() -> str:
    """Generate random phone number (no prefix).

    Returns:
        str: Random phone number.
    """
    return "".join([str(x) for x in choices(range(9), k=9)])


def generate_email() -> str:
    """Generate random e-mail address.

    Returns:
        str: Random e-mail address.
    """
    return generate_str() + "@" + generate_str() + ".com"


def generate_lang_exp():
    """Create vector of language exposures.

    Returns:
        list[float]: Vector of language exposures that adds up to 100.
    """
    nlangs = choice(range(1, 4))
    exp = [0] * 4
    for lang in range(1, nlangs):
        exp[lang] = choice(range(100 - sum(exp)))
    exp[-1] = 100 - sum(exp[:-1])
    exp.sort(reverse=True)
    return exp


def create_record_ppt(is_new: bool = True) -> dict:
    """Create a REDCap participant record.

    Args:
        is_new (bool): Should a new record be created? Defaults to True.

    Returns:
        dict: A REDCap record.
    """
    ppt_id = choice(list(RECORDS.participants.records.keys()))
    return {
        "record_id": api.get_next_id() if is_new else ppt_id,
        "participant_date_created": datetime(2024, 12, 16, 11, 13, 0),
        "participant_date_updated": datetime(2024, 12, 16, 11, 13, 0),
        "participant_name": generate_str(),
        "participant_age_created_months": choice(range(12)),
        "participant_age_created_days": choice(range(31)),
        "participant_sex": str(choice(range(1, 6))),
        "participant_source": str(choice(range(1, 3))),
        "participant_twin": "",
        "participant_parent1_name": generate_str(),
        "participant_parent1_surname": generate_str(),
        "participant_isdropout": choice(["0", "1"]),
        "participant_email1": generate_email(),
        "participant_phone1": generate_phone(),
        "participant_parent2_name": generate_str(),
        "participant_parent2_surname": generate_str(),
        "participant_email2": generate_email(),
        "participant_phone2": generate_str(),
        "participant_address": generate_str(),
        "participant_city": generate_str(),
        "participant_postcode": "".join([str(x) for x in choices(range(9), k=5)]),
        "participant_birth_type": choice(["1", "2"]),
        "participant_gest_weeks": str(choice(range(34, 43))),
        "participant_birth_weight": str(choice(range(2700, 4500))),
        "participant_head_circumference": str(choice(range(32, 38))),
        "participant_apgar1": str(choice(range(10))),
        "participant_apgar2": str(choice(range(10))),
        "participant_apgar3": str(choice(range(10))),
        "participant_hearing": choice(["1", "2"]),
        "participant_diagnoses": generate_str(50),
        "participant_comments": " ".join([generate_str(25) for _ in range(3)]),
        "participants_complete": "2",
    }


def create_record_apt(is_new: bool = True) -> dict:
    """Create a REDCap appointment record.

    Args:
        is_new (bool): Should a new record be created? Defaults to True.

    Returns:
        dict: A REDCap record.
    """
    ppd_id_list = list(RECORDS.participants.records.keys())
    ppt_id = choice(ppd_id_list)
    apt_recs = RECORDS.participants.records[ppt_id].appointments.records

    while not apt_recs:
        ppt_id = choice(ppd_id_list)
        apt_recs = RECORDS.participants.records[ppt_id].appointments.records

    apt_id = choice(list(apt_recs.keys()))

    return {
        "record_id": ppt_id,
        "redcap_repeat_instrument": "appointments",
        "redcap_repeat_instance": (
            api.get_next_id() if is_new else apt_id.split(":")[1]
        ),
        "appointment_study": choice(list(DATA_DICT["appointment_study"].keys())),
        "appointment_date_created": datetime(2024, 12, 12, 14, 9, 0),
        "appointment_date_updated": datetime(2024, 12, 14, 12, 8, 0),
        "appointment_date": datetime(2024, 12, 31, 14, 9, 0),
        "appointment_transport": choice(
            list(DATA_DICT["appointment_transport"].keys())
        ),
        "appointment_taxi_address": generate_str(),
        "appointment_taxi_isbooked": choice(["0", "1"]),
        "appointment_status": choice(list(DATA_DICT["appointment_status"].keys())),
        "appointment_comments": ". ".join([generate_str(25) for _ in range(3)]),
        "appointments_complete": "2",
    }


def create_record_que(is_new: bool = True) -> dict:
    """Create a REDCap questionnaire record.

    Args:
        is_new (bool): Should a new record be created? Defaults to True.

    Returns:
        dict: A REDCap record.
    """
    lang_exp = generate_lang_exp()
    ppt_id_list = list(RECORDS.participants.records.keys())
    ppt_id = choice(ppt_id_list)
    que_recs = RECORDS.participants.records[ppt_id].questionnaires.records

    while not que_recs:
        ppt_id = choice(ppt_id_list)
        que_recs = RECORDS.participants.records[ppt_id].questionnaires.records

    que_id = choice(list(que_recs.keys()))

    date = datetime(2024, 12, 12, 14, 24, 0)

    return {
        "record_id": ppt_id,
        "redcap_repeat_instrument": "language",
        "redcap_repeat_instance": (
            api.get_next_id() if is_new else que_id.split(":")[1]
        ),
        "language_date_created": date,
        "language_date_updated": date,
        "language_isestimated": choice(["0", "1"]),
        "language_lang1": choice(list(DATA_DICT["language_lang1"].keys())),
        "language_lang1_exp": lang_exp[0],
        "language_lang2": choice(list(DATA_DICT["language_lang2"].keys())),
        "language_lang2_exp": lang_exp[1],
        "language_lang3": choice(list(DATA_DICT["language_lang3"].keys())),
        "language_lang3_exp": lang_exp[2],
        "language_lang4": choice(list(DATA_DICT["language_lang4"].keys())),
        "language_lang4_exp": lang_exp[3],
        "language_comments": "",
        "language_complete": "2",
    }



"""Test util functions"""

from datetime import date, datetime
from random import choice, sample
from typing import Generator, Sequence

import polars as pl
from pytest import raises

from babylab import api, utils
from tests import conftest


def test_fmt_labels_dict():
    x = {
        "source": "1",
        "sex": "2",
        "isdropout": "0",
        "age_now_months": "2",
        "birth_type": None,
    }

    o = api.fmt_labels(x)

    assert isinstance(o, dict)
    assert all(k in o for k in x)
    assert isinstance(o["source"], str)
    assert isinstance(o["sex"], str)
    assert isinstance(o["isdropout"], bool)
    assert isinstance(o["age_now_months"], int)

    assert o["source"] == "SJD Àrea de la Dona"
    assert o["sex"] == "Male"
    assert o["isdropout"] is False
    assert o["age_now_months"] == 2
    assert o["birth_type"] is None


def test_fmt_labels_polars():
    x = pl.DataFrame(
        data={
            "source": "1",
            "sex": "2",
            "isdropout": "0",
            "age_now_months": "2",
            "birth_type": None,
        },
    )

    o = api.fmt_labels(x)

    assert isinstance(o, pl.DataFrame)
    assert all(k in o for k in x.columns)

    assert o["source"][0] == "SJD Àrea de la Dona"
    assert o["sex"][0] == "Male"
    assert not o["isdropout"][0]
    assert o["age_now_months"][0] == 2
    assert o["birth_type"][0] is None


def test_get_ppt_table():
    df = utils.get_ppt_table(conftest.RECORDS)

    assert isinstance(df, pl.DataFrame)
    assert all(c in df.columns for c in utils.COLNAMES["participants"])


def test_get_ppt_table_study():
    for k in conftest.DATA_DICT["appointment_study"].items():
        df = utils.get_ppt_table(conftest.RECORDS, study=k)

        assert isinstance(df, pl.DataFrame)
        assert all(c in df.columns for c in utils.COLNAMES["participants"])


def test_get_ppt_table_id_list(ppt_id: str | list[str] | None = None, k: int = 100):
    if ppt_id is None:
        ppt_id = list(conftest.RECORDS.participants.records.keys())

    if isinstance(ppt_id, str):
        ppt_id = [ppt_id]

    df = utils.get_ppt_table(conftest.RECORDS, ppt_id=sample(ppt_id, k=k))

    assert isinstance(df, pl.DataFrame)
    assert all(c in df.columns for c in utils.COLNAMES["participants"])
    assert all(p in ppt_id for p in df["record_id"].unique().to_list())


def test_get_apt_table():
    df = utils.get_apt_table(conftest.RECORDS)

    assert isinstance(df, pl.DataFrame)


def test_get_apt_table_study():
    for k, v in conftest.DATA_DICT["appointment_study"].items():
        df = utils.get_apt_table(conftest.RECORDS, study=k)

        assert isinstance(df, pl.DataFrame)
        assert all(df["study"] == v)


def test_get_apt_table_id(ppt_id: str | None = None):
    if ppt_id is None:
        ppt_id = choice(list(conftest.RECORDS.participants.records.keys()))

    df = utils.get_apt_table(conftest.RECORDS, ppt_id=ppt_id)

    assert isinstance(df, pl.DataFrame)


def test_get_apt_table_id_list(apt_id: str | Sequence[str] | None = None, k: int = 100):
    if apt_id is None:
        apt_id = list(conftest.RECORDS.appointments.records.keys())

    if isinstance(apt_id, str):
        apt_id = [apt_id]

    ppt_id = set(i.split(":")[0] for i in apt_id)

    if k > len(ppt_id):
        k = len(ppt_id)

    df = utils.get_apt_table(conftest.RECORDS, ppt_id=sample(list(ppt_id), k=k))

    assert isinstance(df, pl.DataFrame)
    assert all(p in ppt_id for p in df["record_id"].unique().to_list())


def test_get_que_table():
    df = utils.get_que_table(conftest.RECORDS)

    assert isinstance(df, pl.DataFrame)


def test_get_que_table_id(ppt_id: str | list[str] | None = None):
    if ppt_id is None:
        ppt_id = choice(list(conftest.RECORDS.participants.records.keys()))

    df = utils.get_que_table(conftest.RECORDS, ppt_id=ppt_id)

    assert isinstance(df, pl.DataFrame)


def test_get_que_table_id_list(que_id: str | list[str] | None = None, k: int = 100):
    if que_id is None:
        que_id = list(conftest.RECORDS.appointments.records.keys())

    if isinstance(que_id, str):
        que_id = [que_id]

    ppt_id = set(i.split(":")[0] for i in que_id)

    if k > len(ppt_id):
        k = len(ppt_id)

    df = utils.get_apt_table(conftest.RECORDS, ppt_id=sample(list(ppt_id), k=k))
    assert isinstance(df, pl.DataFrame)
    assert all(p in ppt_id for p in df["record_id"].unique().to_list())


def test_is_in_data_dict():
    """Test is_in_datadict."""

    assert utils.is_in_data_dict("appointment_status", ["Successful"]) == ["Successful"]

    assert utils.is_in_data_dict(
        "appointment_status",
        ["Successful", "Confirmed"],
    ) == [
        "Successful",
        "Confirmed",
    ]

    assert utils.is_in_data_dict("appointment_status", "Successful") == ["Successful"]

    assert utils.is_in_data_dict("appointment_study", ["mop_newborns_1_nirs"]) == [
        "mop_newborns_1_nirs"
    ]

    assert utils.is_in_data_dict(
        "appointment_study", ["mop_newborns_1_nirs", "mop_infants_1_hpp"]
    ) == ["mop_newborns_1_nirs", "mop_infants_1_hpp"]

    assert utils.is_in_data_dict("appointment_study", "mop_newborns_1_nirs") == [
        "mop_newborns_1_nirs"
    ]

    with raises(ValueError):
        utils.is_in_data_dict("appointment_status", ["Badname"])
        utils.is_in_data_dict("appointment_status", ["Badname", "Successful"])
        utils.is_in_data_dict("appointment_status", "Badname")


def test_get_year_weeks():
    """Test get_year_weeks."""
    assert isinstance(utils.get_year_weeks(2025), Generator)
    assert isinstance(next(utils.get_year_weeks(2025)), date)


def test_get_week_n():
    """Test get_week_n."""
    assert isinstance(utils.get_week_n(datetime.today()), int)


def test_get_weekly_apts():
    """Test get_weekly_apts."""
    assert isinstance(
        utils.get_weekly_apts(records=conftest.RECORDS),
        int,
    )
    assert isinstance(
        utils.get_weekly_apts(records=conftest.RECORDS, study="mop_newborns_1_nirs"),
        int,
    )
    assert isinstance(
        utils.get_weekly_apts(
            records=conftest.RECORDS, study="mop_newborns_1_nirs", status="Successful"
        ),
        int,
    )
    assert isinstance(
        utils.get_weekly_apts(
            records=conftest.RECORDS,
            study=["mop_newborns_1_nirs", "mop_infants_1_hpp"],
        ),
        int,
    )
    assert isinstance(
        utils.get_weekly_apts(
            records=conftest.RECORDS,
            study=["mop_newborns_1_nirs", "mop_infants_1_hpp"],
            status=["Successful", "Confirmed"],
        ),
        int,
    )



"""Global variables."""

from polars import Boolean, Datetime, Int64, Schema, String

URI = "https://apps.sjdhospitalbarcelona.org/redcap/api/"

COLNAMES = {
    "participants": [
        "record_id",
        "date_created",
        "date_updated",
        "source",
        "name",
        "age_created_months",
        "age_created_days",
        "sex",
        "twin",
        "isdropout",
        "parent1_name",
        "parent1_surname",
        "email1",
        "phone1",
        "parent2_name",
        "parent2_surname",
        "email2",
        "phone2",
        "address",
        "city",
        "postcode",
        "birth_type",
        "gest_weeks",
        "birth_weight",
        "head_circumference",
        "apgar1",
        "apgar2",
        "apgar3",
        "hearing",
        "diagnoses",
        "comments",
        "age_now_months",
        "age_now_days",
    ],
    "appointments": [
        "record_id",
        "appointment_id",
        "study",
        "status",
        "date",
        "date_created",
        "date_updated",
        "transport",
        "taxi_address",
        "taxi_isbooked",
        "confirmation_comments",
        "comments",
    ],
    "questionnaires": [
        "record_id",
        "questionnaire_id",
        "date_created",
        "date_updated",
        "isestimated",
        "lang1",
        "lang1_exp",
        "lang2",
        "lang2_exp",
        "lang3",
        "lang3_exp",
        "lang4",
        "lang4_exp",
        "comments",
        "complete",
    ],
}

FIELDS_TO_RENAME = [
    "participant_sex",
    "participant_birth_type",
    "participant_hearing",
    "participant_source",
    "appointment_study",
    "appointment_status",
    "appointment_transport",
    "language_lang1",
    "language_lang2",
    "language_lang3",
    "language_lang4",
]


INT_FIELDS = [
    "age_created_months",
    "age_created_days",
    "age_now_months",
    "age_now_days",
    "gest_weeks",
    "birth_weight",
    "head_circumference",
    "apgar1",
    "apgar2",
    "apgar3",
]

SCHEMA = {
    "participants": Schema(
        [
            ("record_id", String),
            ("date_created", Datetime(time_unit="us", time_zone=None)),
            ("date_updated", Datetime(time_unit="us", time_zone=None)),
            ("source", String),
            ("is_born", Boolean),
            ("name", String),
            ("age_created_months", Int64),
            ("age_created_days", Int64),
            ("sex", String),
            ("twin", String),
            ("isdropout", Boolean),
            ("parent1_name", String),
            ("parent1_surname", String),
            ("email1", String),
            ("phone1", String),
            ("parent2_name", String),
            ("parent2_surname", String),
            ("email2", String),
            ("phone2", String),
            ("address", String),
            ("city", String),
            ("postcode", String),
            ("birth_type", String),
            ("gest_weeks", Int64),
            ("birth_weight", Int64),
            ("head_circumference", Int64),
            ("apgar1", Int64),
            ("apgar2", Int64),
            ("apgar3", Int64),
            ("hearing", String),
            ("diagnoses", String),
            ("comments", String),
            ("age_now_months", Int64),
            ("age_now_days", Int64),
        ]
    ),
    "appointments": Schema(
        [
            ("record_id", String),
            ("redcap_repeat_instance", Int64),
            ("study", String),
            ("date_created", Datetime(time_unit="us", time_zone=None)),
            ("date_updated", Datetime(time_unit="us", time_zone=None)),
            ("date", Datetime(time_unit="us", time_zone=None)),
            ("transport", String),
            ("taxi_address", String),
            ("taxi_isbooked", Boolean),
            ("status", String),
            ("confirmation_comments", String),
            ("comments", String),
            ("id", String),
        ]
    ),
    "questionnaires": Schema(
        [
            ("record_id", String),
            ("redcap_repeat_instance", Int64),
            ("date_created", Datetime(time_unit="us", time_zone=None)),
            ("date_updated", Datetime(time_unit="us", time_zone=None)),
            ("isestimated", Boolean),
            ("lang1", String),
            ("lang1_exp", Int64),
            ("lang2", String),
            ("lang2_exp", Int64),
            ("lang3", String),
            ("lang3_exp", Int64),
            ("lang4", String),
            ("lang4_exp", Int64),
            ("comments", String),
            ("complete", String),
        ]
    ),
}



"""Test API."""

import os
from datetime import datetime

import pytest

from babylab import api
from tests import conftest

IS_GIHTUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"


def test_get_api_key():
    """Test get_api_key."""
    token = api.get_api_key()
    assert token is not None
    assert isinstance(token, str)

    if not IS_GIHTUB_ACTIONS:
        token = api.get_api_key(path="~/.env")
        assert token is not None
        assert isinstance(token, str)


def test_redcap_version():
    """Test ``redcap_version``."""
    version = api.get_redcap_version()
    assert version
    assert isinstance(version, str)
    assert len(version.split(".")) == 3


def test_dt_to_str():
    """Test ``test_datetimes_to_str`` function."""
    data = {
        "date_now": datetime(2024, 10, 24, 8, 48, 34, 685496),
        "date_today": datetime(2024, 10, 24, 8, 48),
        "date_str": "2024-05-12T5:12",
    }
    result = api.dt_to_str(data)
    assert result["date_now"] == "2024-10-24T08:48:34.685496"
    assert result["date_today"] == "2024-10-24T08:48:00"
    assert result["date_str"] == data["date_str"]


def test_str_to_dt():
    """Test ``test_datetimes_to_str`` function."""
    data = {
        "date_now": "2024-05-12 05:34:15",
        "date_today": "2024-05-12 05:34",
    }
    result = api.str_to_dt(data)
    assert result["date_now"] == datetime(2024, 5, 12, 5, 34, 15)
    assert result["date_today"] == datetime(2024, 5, 12, 5, 34)


def test_get_data_dict():
    """Test ``get_records``."""
    data_dict = api.get_data_dict()
    assert isinstance(data_dict, dict)
    assert all(isinstance(v, dict) for v in data_dict.values())


def test_make_id():
    """Test ``make_id``"""
    assert api.make_id(1, 2) == "1:2"
    assert api.make_id(1) == "1"
    assert api.make_id("1", "2") == "1:2"
    assert api.make_id("1") == "1"
    with pytest.raises(ValueError):
        api.make_id(1, "a")
    with pytest.raises(ValueError):
        api.make_id(1, "1 ")
    with pytest.raises(ValueError):
        api.make_id("a")
    with pytest.raises(ValueError):
        api.make_id("1 ")
    with pytest.raises(ValueError):
        api.make_id("1:1")


def test_get_records():
    """Test ``get_records``."""
    records = api.get_records()
    assert isinstance(records, list)
    assert all(isinstance(r, dict) for r in records)


def test_add_participant():
    """Test adding,modifying, and deleting participants."""
    ppt_record = conftest.create_record_ppt(is_new=True)

    rid = ppt_record["record_id"]
    api.add_participant(ppt_record)
    assert isinstance(api.get_participant(rid), api.Participant)

    old_val = ppt_record["participant_apgar1"]
    new_val = int(old_val) + 1 if old_val != 10 else 1
    ppt_record["participant_apgar1"] = new_val
    api.add_participant(ppt_record, modifying=True)
    new_record = api.get_participant(rid)
    assert new_record.data["apgar1"] == new_val


def test_add_appointment():
    """Test adding appointments."""
    apt_record = conftest.create_record_apt(is_new=True)

    api.add_appointment(apt_record)
    rid = api.make_id(apt_record["record_id"], apt_record["redcap_repeat_instance"])
    assert isinstance(api.get_appointment(rid), api.Appointment)

    old_val = apt_record["appointment_status"]
    new_val = "1" if old_val != "1" else "1"
    apt_record["appointment_status"] = new_val
    api.add_appointment(apt_record)
    new_record = api.get_appointment(rid)
    new_val_label = conftest.DATA_DICT["appointment_status"][new_val]
    assert new_record.data["status"] == new_val_label


def test_add_questionnaire():
    """Test adding questionnaires."""
    que_record = conftest.create_record_que(is_new=True)

    api.add_questionnaire(que_record)
    rid = api.make_id(que_record["record_id"], que_record["redcap_repeat_instance"])
    assert isinstance(api.get_questionnaire(rid), api.Questionnaire)

    old_val = que_record["language_lang1"]
    new_val = str(1 if old_val == "0" else 0)
    que_record["language_lang1"] = new_val
    api.add_questionnaire(que_record)
    new_record = api.get_questionnaire(rid)
    new_val_label = conftest.DATA_DICT["language_lang1"][new_val]
    assert new_record.data["lang1"] == new_val_label


def test_parse_str_date():
    """Test parse_str_date."""
    hms = datetime(2025, 5, 2, 9, 10, 10)
    hm = datetime(2025, 5, 2, 9, 10)
    assert api.parse_str_date("2025-05-02T09:10:10") == hms
    assert api.parse_str_date("2025-05-02 09:10:10") == hms
    assert api.parse_str_date("2025-05-02 09:10") == hm


def test_get_age():
    """Test ``get_age``"""
    ts = datetime(2024, 5, 1, 3, 4)
    ts_new = datetime(2025, 1, 4, 1, 2)
    age = (5, 5)

    # when only birth date is provided
    assert isinstance(api.get_age(age, ts), tuple)
    assert all(isinstance(d, int) for d in api.get_age(age, ts))
    assert len(api.get_age(age, ts)) == 2

    # when birth date AND ts_new are provided
    assert isinstance(api.get_age(age, ts, ts_new=ts_new), tuple)
    assert all(isinstance(d, int) for d in api.get_age(age, ts, ts_new))
    assert len(api.get_age(age, ts, ts_new)) == 2

    assert api.get_age(age, ts, ts_new) == (13, 7)

    assert all(d > 0 for d in api.get_age(age, ts, ts_new))
    with pytest.raises(api.BadAgeFormat):
        api.get_age(age="5, 4", ts=ts)  # ty: ignore[invalid-argument-type]



"""
Fixtures for testing.
"""

from babylab import api
from tests import conftest

records = api.Records()
data_dict = api.get_data_dict()
ppt = conftest.create_record_ppt()
apt = conftest.create_record_apt()
que = conftest.create_record_que()



"""babylab-redcap

A Python interface to the for the SJD Babylab REDCap database API.


This package provides convenient wrappers to interact with the NeuroDevCo Babylab's REDCap API to consult, modify or delete records from participants, appointments, or language questionnsires.

Modules:
    api: Contains the main functions used to interact with the REDCap API.
    utils: Utils functions for cleaning data, and for reformatting labels and values.
    globals: Global variables used throughout the `api` and `utiils` modules.
"""  # pylint: disable=line-too-long



"""
Util functions for the app.
"""

from collections.abc import Sequence
from copy import deepcopy
from datetime import date, datetime, timedelta
from typing import Generator

import polars as pl

from babylab import api
from babylab.globals import COLNAMES


def is_in_data_dict(variable: str, x: Sequence[str] | str | None = None) -> list[str]:
    """Check that a value is an element in the data dictionary.

    Args:
        variable (str): Key in which to look for.
        x (Sequence[str] | str | None, optional): Value to look up in the data dictionary. Defaults to None (all options are checked for the variable).

    Raises:
        ValueError: If `x` is not an option present in `data_dict`.

    Returns:
        list[str]: Values in data dict.
    """
    options = list(api.DATA_DICT[variable].values())

    if x is None:
        return options

    x = [x] if isinstance(x, str) else list(x)

    for xi in x:
        if xi not in options:
            raise ValueError(f"{xi} is not an option in {variable}")

    return x


def get_ppt_table(
    records: api.Records,
    ppt_id: list[str] | str | None = None,
    study: list[str] | str | None = None,
) -> pl.DataFrame:
    """Get participants table

    Args:
        records (api.Records): REDCap records, as returned by ``api.Records``.
        ppt_id (list[str] | str | None): ID of participant to return. If None (default), all participants are returned.
        study (list[str] | str | None, optional): Study in which the participant in the records must have participated to be kept. Defaults to None.

    Returns:
        pl.DataFrame: Table of partcicipants.
    """
    if not records.participants.records:
        return pl.DataFrame(schema=COLNAMES["participants"])

    if isinstance(ppt_id, str):
        ppt_id = [ppt_id]

    if isinstance(study, str):
        study = [study]

    df = api.to_df(records.participants)

    if study:
        ppt_study = (
            api.to_df(records.appointments)
            .filter(pl.col("study").is_in(study))
            .unique("record_id")
            .get_column("record_id")
            .to_list()
        )
        df = df.filter(pl.col("record_id").is_in(ppt_study))

    if ppt_id:
        df = df.filter(pl.col("record_id").is_in(ppt_id))

    return df


def get_apt_table(
    records: api.Records,
    ppt_id: list[str] | str | None = None,
    study: list[str] | str | None = None,
) -> pl.DataFrame:
    """Get appointments table.

    Args:
        records (api.Records): REDCap records, as returned by ``api.Records``.
        ppt_id (list[str] | str | None, optional): ID of participant to return. If None (default), all participants are returned.
        study (list[str] | str | None, optional): Study to filter for. If None (default) all studies are returned.

    Returns:
        DataFrame: Table of appointments.
    """
    df = api.to_df(deepcopy(records.appointments))

    if len(df) == 0:
        return pl.DataFrame(schema=COLNAMES["appointments"])

    if isinstance(study, str):
        study = [study]

    if isinstance(ppt_id, str):
        ppt_id = [ppt_id]

    if study:
        df = df.filter(pl.col("study").is_in(study))

    if ppt_id:
        df = df.filter(pl.col("record_id").is_in(ppt_id))

    ppt_df = api.to_df(records.participants)

    cols = ["record_id", "age_now_months", "age_now_days"]
    df = df.join(ppt_df.select(cols), on="record_id")
    _, col_1, col_2 = cols

    months, days = [], []
    for rows in df.iter_rows(named=True):
        t = datetime.now() if rows["date"] is None else rows["date"]
        m, d = api.get_age(age=(rows[col_1], rows[col_2]), ts=t)
        months.append(m)
        days.append(d)

    df.insert_column(-1, pl.Series("age_apt_months", months))
    df.insert_column(-1, pl.Series("age_apt_days", days))

    return df


def get_que_table(
    records: api.Records, ppt_id: list[str] | str | None = None
) -> pl.DataFrame:
    """Get questionnaires table.

    Args:
        records (api.Records): REDCap records, as returned by ``api.Records``.
        ppt_id (list[str] | str | None, optional): ID of participant to return. If None (default), all participants are returned.
        relabel (bool): Reformat labels if True (default).

    Returns:
        DataFrame: A formated Pandas DataFrame.
    """
    df = api.to_df(deepcopy(records.questionnaires))

    if len(df) == 0:
        return pl.DataFrame(schema=COLNAMES["questionnaires"])

    if isinstance(ppt_id, str):
        ppt_id = [ppt_id]

    if ppt_id:
        df = df.filter(pl.col("record_id").is_in(ppt_id))

    return df


def count_col(
    x: pl.DataFrame,
    col: str,
    values_sort: bool = False,
    cumulative: bool = False,
    missing_label: str = "Missing",
) -> dict:
    """Count frequencies of column in DataFrame.

    Args:
        x (DataFrame): DataFrame containing the target column.
        col (str): Name of the column.
        values_sort (str, optional): Should the resulting dict be ordered by values? Defaults to False.
        cumulative (bool, optional): Should the counts be cumulative? Defaults to False.
        missing_label (str, optional): Label to associate with missing values. Defaults to "Missing".

    Returns:
        dict: Counts of each category, sorted in descending order.
    """
    counts = x[col].value_counts().to_dict()
    counts = {missing_label if not k else k: v for k, v in counts.items()}

    if values_sort:
        counts = dict(sorted(counts.items(), key=lambda item: item[1], reverse=True))

    if cumulative:
        cumsum = 0
        for key in counts:
            cumsum += counts[key]
            counts[key] = cumsum

    return counts


def get_year_weeks(year: int) -> Generator[datetime, datetime, datetime]:
    """Get week numbers of the year.

    Args:
        year (int): Year to get weeks for.

    Yields:
        Generator[datetime, datetime, datetime]: Number of weeks in the year.
    """
    date_first = date(year, 1, 1)
    date_first += timedelta(days=6 - date_first.weekday())

    while date_first.year == year:
        yield date_first
        date_first += timedelta(days=7)


def get_week_n(timestamp: date) -> int:
    """Get current week number in the year.

    Args:
        timestamp (date): Date to calculate week number for.

    Returns:
        int: Week number of given date.
    """
    weeks = {}

    for wn, d in enumerate(get_year_weeks(timestamp.year)):
        weeks[wn + 1] = [(d + timedelta(days=k)).isoformat() for k in range(7)]

    for k, v in weeks.items():
        if datetime.strftime(timestamp, "%Y-%m-%d") in v:
            return k
    return k


def get_weekly_apts(
    records: api.Records,
    study: Sequence | str | None = None,
    status: Sequence | str | None = None,
) -> int:
    """Get weekly number of appointments.

    Args:
        records (api.Records): REDCap records, as returned by ``api.Records``.
        study (list | None, optional): Study to filter for. Defaults to None.
        status (list | None, optional): Status to filter for. Defaults to None.

    Returns:
        int: Weekly number of appointment with for a given study and/or status.

    Raises:
        ValueError: If `study` or `status` is not available.
    """
    study = is_in_data_dict("appointment_study", study)
    status = is_in_data_dict("appointment_status", status)
    apts = records.appointments.records.values()

    date = get_week_n(datetime.today())

    return sum(
        get_week_n(v.data["date_created"]) == date
        for v in apts
        if v.data["status"] in status and v.data["study"] in study
    )



"""Test database models."""

from datetime import datetime

from polars import DataFrame

from babylab import api
from tests import conftest

PPT_RECORD = conftest.create_record_ppt(is_new=True)
APT_RECORD = conftest.create_record_apt(is_new=True)
QUE_RECORD = conftest.create_record_que(is_new=True)


def test_participant_class():
    """Test participant class."""
    data = api.prepare_data(PPT_RECORD)
    x = api.Participant(data["record_id"], data)

    for att in ["ppt_id", "data"]:
        assert hasattr(x, att)

    assert isinstance(x.ppt_id, str)
    assert isinstance(x.data, dict)


def test_appointment_class():
    """Test appointment class."""
    data = api.prepare_data(APT_RECORD, "apt")
    x = api.Appointment(data["record_id"], data)

    for att in ["ppt_id", "apt_id", "date", "status", "data"]:
        assert hasattr(x, att)

    assert isinstance(x.ppt_id, str)
    assert isinstance(x.apt_id, str)
    assert isinstance(x.date, datetime)
    assert isinstance(x.status, str)
    assert isinstance(x.data, dict)


def test_questionnaire_class():
    """Test questionnaire class."""
    data = api.prepare_data(QUE_RECORD, "que")
    x = api.Questionnaire(data["record_id"], data)

    for att in ["ppt_id", "que_id", "isestimated", "data"]:
        assert hasattr(x, att)

    assert isinstance(x.ppt_id, str)
    assert isinstance(x.que_id, str)
    assert isinstance(x.isestimated, bool)
    assert isinstance(x.data, dict)


def test_records_class():
    """Test participant class."""
    assert hasattr(conftest.RECORDS, "appointments")
    assert hasattr(conftest.RECORDS, "participants")
    assert hasattr(conftest.RECORDS, "questionnaires")

    assert isinstance(conftest.RECORDS.appointments, api.RecordList)
    assert isinstance(conftest.RECORDS.participants, api.RecordList)
    assert isinstance(conftest.RECORDS.questionnaires, api.RecordList)

    assert isinstance(repr(conftest.RECORDS), str)
    assert "REDCap database" in repr(conftest.RECORDS)
    assert isinstance(str(conftest.RECORDS), str)
    assert "REDCap database" in str(conftest.RECORDS)


def test_recordlist_class_participants():
    """Test RecordList class with participants."""
    records = conftest.RECORDS.participants
    assert isinstance(records.records, dict)
    assert isinstance(api.to_df(records), DataFrame)
    assert isinstance(records.kind, str)
    assert records.kind == "participants"


def test_recordlist_class_appointments():
    """Test RecordList class with appointments."""
    records = conftest.RECORDS.appointments

    assert isinstance(records.records, dict)
    assert isinstance(api.to_df(records), DataFrame)
    assert isinstance(records.kind, str)
    assert records.kind == "appointments"


def test_recordlist_class_questionnaires():
    """Test RecordList class with questionnaires."""
    records = conftest.RECORDS.questionnaires

    assert isinstance(records.records, dict)
    assert isinstance(api.to_df(records), DataFrame)
    assert isinstance(records.kind, str)
    assert records.kind == "questionnaires"


def test_records_class_participants():
    """Test records class (Participants)"""
    assert hasattr(conftest.RECORDS.participants, "records")
    assert isinstance(conftest.RECORDS.participants.records, dict)
    assert all(
        isinstance(r, api.Participant)
        for r in conftest.RECORDS.participants.records.values()
    )


def test_records_class_appointments():
    """Test records class (Appointments)"""
    assert hasattr(conftest.RECORDS.appointments, "records")
    assert isinstance(conftest.RECORDS.appointments.records, dict)
    assert all(
        isinstance(r, api.Appointment)
        for r in conftest.RECORDS.appointments.records.values()
    )


def test_records_class_questionnaires():
    """Test records class (Questionnaires)"""
    assert hasattr(conftest.RECORDS.questionnaires, "records")
    assert isinstance(conftest.RECORDS.questionnaires.records, dict)
    assert all(
        isinstance(r, api.Questionnaire)
        for r in conftest.RECORDS.questionnaires.records.values()
    )
